/*
 * Assembly Language Debugger
 *
 * Copyright (C) 2000 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 *
 * $Id: c_disassemble.c,v 1.8 2001/04/07 01:00:00 cosine Exp $
 */

#include <stdlib.h>
#include <sys/types.h>
#include <errno.h>
#include <string.h>
#include <assert.h>

#include "alddefs.h"
#include "defs.h"
#include "disassemble.h"
#include "file.h"
#include "load.h"
#include "msg.h"
#include "scroll.h"
#include "print.h"
#include "window.h"

/*
 * libDebug includes
 */
#include "debug.h"

/*
 * libOp includes
 */
#include "disasm.h"

/*
 * libString includes
 */
#include "alloc.h"
#include "Strn.h"

static void ResetDisassemblyPosition();

/*
c_disassemble()
 Disassemble the current file

Format for this command:
  disassemble [start [stop]] [-num <number>] [-section <name>]
*/

void
c_disassemble(int ac, char **av)

{
  unsigned char *data;  /* data to disassemble */
  unsigned char *membuf;
  long insnum,          /* number of instructions to disassemble */
       inscnt;          /* number of instructions disassembled so far */
  char *endptr;
  long length;
  int cnt,
      alen;
  unsigned int start,
               end;
  unsigned int address; /* address of instruction to be disassembled */
  char *section;        /* are we disassembling a specific section? */
  long ndumped,         /* number of bytes dumped */
       numbytes;        /* total bytes */
  unsigned char spill[MAXOPLEN + 1];

  start = end = 0;
  section = 0;
  insnum = 0;
  ndumped = numbytes = 0;

  for (cnt = 1; cnt < ac; ++cnt)
  {
    alen = strlen(av[cnt]);
    if (!Strncasecmp(av[cnt], "-section", alen))
    {
      if (++cnt >= ac)
      {
        Print(P_COMMAND, "No section name specified");
        return;
      }

      /*
       * Check if they already specified a starting address
       */
      if (start)
        continue;

      if (!FindSectionByName(MapPtr, av[cnt], &start, &end))
      {
        Print(P_COMMAND, "No such section: %s", av[cnt]);
        return;
      }

      section = av[cnt];
    }
    else if (!Strncasecmp(av[cnt], "-num", alen))
    {
      if (++cnt >= ac)
      {
        Print(P_COMMAND, "No instruction count specified");
        return;
      }

      insnum = strtol(av[cnt], &endptr, 0);
      if ((endptr == av[cnt]) || (*endptr != '\0'))
      {
        Print(P_COMMAND, MSG_INVNUM, av[cnt]);
        return;
      }
    }
    else
    {
      /*
       * It must be the starting or ending address
       */
      if (!start)
        start = strtoul(av[cnt], &endptr, 0);
      else if (!end)
        end = strtoul(av[cnt], &endptr, 0);

      if ((endptr == av[cnt]) || (*endptr != '\0'))
      {
        Print(P_COMMAND, MSG_INVNUM, av[cnt]);
        return;
      }

      /*
       * Start/End addresses take priority over -section
       */
      section = 0;
    }
  }

  if (section)
  {
    /*
     * We are disassembling a specific section - set
     * ObjectFileOffset to the correct position.
     */
    VirtualObjectFileOffset = start;
    ObjectFileOffset = VirtualObjectFileOffset - VirtualFileAddress;

    Print(P_COMMAND, "Disassembling section %s (0x%08X - 0x%08X)",
      section,
      start,
      end);

  #ifdef USE_CURSES

    /*
     * Clear any disassembly data from previous runs
     */
    if (ModeCurses)
      ClearScrollData(DisassemblyFrame);

  #endif /* USE_CURSES */
  } /* if (section) */

  membuf = 0;
  if (start)
  {
    if (!end)
    {
      /*
       * They gave a starting memory address, but did not say when
       * to stop, so default to 15 instructions
       */
      insnum = 15;

      /*
       * 150 bytes should be sufficient for 15 instructions
       */
      end = start + 150;
    }

    numbytes = end - start;
    if (numbytes < 0)
    {
      Print(P_COMMAND, "Ending address is lower than starting address");
      return;
    }

    /*
     * Add fifteen to account for the very last instruction - in case
     * it starts at location 'end'
     */
    numbytes += 15;

    ndumped = DumpMemory((char **) &membuf, start, numbytes);
    if (!ndumped)
    {
      Print(P_COMMAND, MSG_NOACCESS,
        start,
        strerror(errno));

      if (membuf)
        MyFree(membuf);

      return;
    }

    /*
     * In case we couldn't access all the memory we tried to dump,
     * set 'end' correctly.
     */
    if (ndumped < numbytes)
      end = start + ndumped;

    data = membuf;
    address = start;
  } /* if (start) */
  else
  {
    /*
     * We are using file offsets rather than memory locations
     */

    if (ObjectFileOffset >= MappedSize)
      ResetDisassemblyPosition();

    data = (unsigned char *) MapPtr;
    data += ObjectFileOffset;
    address = VirtualObjectFileOffset;
  }

  inscnt = 0;

  PrintBurstStart();

  /*
   * There are many conditions which could cause the loop to
   * stop, so I'm using an infinite loop with a bunch of 'break'
   * checks.
   */
  while (1)
  {
    /*
     * stop if we reach the instruction limit
     */
    if (insnum && (inscnt++ >= insnum))
      break;

    /*
     * stop if we hit the end of the address range we are
     * disassembling
     */
    if (start)
    {
      if (end && (address > end))
        break;
    }
    else
    {
      /*
       * stop if we go past the end of the file
       */
      if (ObjectFileOffset >= MappedSize)
        break;
      else
      {
        /*
         * Check if we are nearing the end of the file - if so,
         * use a separate buffer rather than MapPtr, because
         * if our disassembling routine screws up, it may try to
         * read a byte or two past the end of MapPtr, causing
         * a crash.
         */
        if ((ObjectFileOffset + MAXOPLEN) >= MappedSize)
        {
          unsigned int ii;

          cnt = 0;
          memset((void *) spill, 0, sizeof(spill));

          for (ii = ObjectFileOffset; ii < MappedSize; ++ii)
            spill[cnt++] = (unsigned char) *((unsigned char *) MapPtr + ii);

          data = spill;
        }
      }
    }

    /*
     * stop if we hit the end of the section we are disassembling
     */
    if (section && (address >= end))
      break;

    /*
     * disassemble the next instruction
     */
    length = Disasm(data, address, P_DISASSEMBLY);

    assert(length != 0);

    if (length > 0)
    {
      data += length;
      address += length;

      if (!membuf)
      {
        ObjectFileOffset += length;
        VirtualObjectFileOffset += length;
      }
    }
    else
      break; /* something went wrong */
  } /* while (1) */

  PrintBurstEnd();

  if (membuf)
  {
    MyFree(membuf);
    if (ndumped < numbytes)
    {
      Print(P_COMMAND, MSG_NOACCESS,
        start + ndumped,
        strerror(errno));
    }
  }

#ifdef USE_CURSES

  if (ModeCurses)
  {
    /*
     * The above Print() statements did not actually output
     * anything to the window - they simply stored the data in
     * the dynamic pointer array DisassemblyFrame->data. We need
     * to do it this way so we can write the data now with the
     * top line highlighted, which DrawWindow() will take care of.
     */
    DrawWindow(DisassemblyFrame);
  }

#endif /* USE_CURSES */
} /* c_disassemble() */

/*
ResetDisassemblyPosition()
  Called when we disassemble the last instruction in the file -
reset the offsets and clear the disassembly window
*/

static void
ResetDisassemblyPosition()

{
  Print(P_COMMAND,
    "End of file reached, repositioning offset to 0x%08x",
    VirtualEntryPoint);

  ObjectFileOffset = EntryPoint;
  VirtualObjectFileOffset = VirtualEntryPoint;

#ifdef USE_CURSES

  /*
   * Clear any disassembly data from previous runs
   */
  if (ModeCurses)
    ClearScrollData(DisassemblyFrame);

#endif /* USE_CURSES */
} /* ResetDisassemblyPosition() */
