/*
 * Assembly Language Debugger
 *
 * Copyright (C) 2000 Patrick Alken
 * This program comes with absolutely NO WARRANTY
 *
 * Should you choose to use and/or modify this source code, please
 * do so under the terms of the GNU General Public License under which
 * this program is distributed.
 *
 * $Id: command.c,v 1.19 2001/04/07 20:22:16 cosine Exp $
 */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <setjmp.h>
#include <signal.h>

#include "alddefs.h"
#include "command.h"
#include "defs.h"
#include "input.h"
#include "load.h"
#include "misc.h"
#include "msg.h"
#include "print.h"
#include "readln.h"
#include "signals.h"
#include "wgets.h"
#include "window.h"

/*
 * libOp includes
 */
#include "disasm.h"

/*
 * libString includes
 */
#include "alloc.h"
#include "Strn.h"

/*
 * Global: command prompt string
 */
char                        *CmdPrompt = 0;

/*
 * Global: environment for GetCommandInput()
 */
jmp_buf                     CmdParserEnv;

static struct Command maincmds[] = {
  /*
   * Aliases - put these at the top since they are used
   * frequently
   */
  { "c", c_continue, C_ALIAS|C_FILELOADED },
  { "d", c_disassemble, C_ALIAS|C_FILELOADED },
  { "e", c_examine, C_ALIAS|C_FILELOADED },
  { "n", c_next, C_ALIAS|C_FILELOADED },
  { "r", c_run, C_ALIAS|C_FILELOADED },
  { "s", c_step, C_ALIAS|C_FILELOADED },

  { "break", c_break, C_FILELOADED },
  { "clear", c_clear, C_CURSES },
  { "continue", c_continue, C_FILELOADED },
  { "delete", c_delete, 0 },
  { "disable", c_disable, 0 },
  { "disassemble", c_disassemble, C_FILELOADED },
  { "dump", c_examine, C_ALIAS|C_FILELOADED },
  { "enable", c_enable, 0 },
  { "enter", c_enter, C_FILELOADED },
  { "examine", c_examine, C_FILELOADED },
  { "exit", c_quit, C_ALIAS },
  { "file", c_file, C_FILELOADED },
  { "help", c_help, 0 },
  { "ignore", c_ignore, 0 },
  { "lbreak", c_lbreak, 0 },
  { "load", c_load, 0 },
  { "next", c_next, C_FILELOADED },
  { "quit", c_quit, 0 },
  { "register", c_register, C_FILELOADED },
  { "run", c_run, C_FILELOADED },
  { "set", c_set, 0 },
  { "step", c_step, C_FILELOADED },
  { "store", c_enter, C_ALIAS|C_FILELOADED },
  { "tbreak", c_tbreak, C_FILELOADED },
  { "window", c_window, C_CURSES },
  { "write", c_write, C_CURSES },

  { 0, 0, 0 }
};

void
ProcessCommand(char *command)

{
  int ac;
  char **av;
  struct Command *cptr;
  unsigned int flags;

  assert(command != 0);

  ac = SplitBuffer(command, &av);

  if (!ac)
    return;

  if (ac == (-1))
  {
    Print(P_COMMAND, "Unbalanced quotes");
    return;
  }

  flags = 0;
  cptr = FindCommand(maincmds, av[0], &flags);

  if (!cptr)
  {
    if (flags & C_AMBIGUOUS)
    {
      Print(P_COMMAND, "Ambiguous command: %s",
        av[0]);
    }
    else
    {
      Print(P_COMMAND, "Unknown command: %s (help shows a list)",
        av[0]);
    }

    MyFree(av);
    return;
  }

  if ((cptr->flags & C_FILELOADED) && !MapPtr)
  {
    Print(P_COMMAND, MSG_NOFILELOADED);
    MyFree(av);
    return;
  }

#ifdef USE_CURSES

  if (!ModeCurses && (cptr->flags & C_CURSES))
  {
    Print(P_COMMAND,
      "The command \"%s\" works only in curses mode",
      cptr->cmd);
    MyFree(av);
    return;
  }

#endif /* USE_CURSES */

  /*
   * Execute the command
   */
  (*cptr->funcptr)(ac, av);

  MyFree(av);
} /* ProcessCommand() */

/*
GetCommandInput()
  Enter infinite loop to wait for and process user commands
*/

void
GetCommandInput()

{
  char buffer[MAXLINE + 1];
  char last[MAXLINE + 1]; /* last command entered */
  char *bufptr,
       *cmdptr;
  int GotLongJmp;

  *last = '\0';

  if (ModeConsole)
  {
    while (1)
    {
      GotLongJmp = setjmp(CmdParserEnv);
      if (GotLongJmp)
      {
        /*
         * In case we were in the middle of a print burst and the
         * user ^C'd, clear the burst flag
         */
        PrintBurstEnd();
      }

      /*
       * This must be set up after the setjmp() call or thing will
       * be very bad if ^C is pressed beforehand
       */
      signal(SIGINT, SigHandler);

      /*
       * Read a line from stdin
       */
      bufptr = ReadLine();
      assert(bufptr != 0);

      if ((*bufptr == '\0') || (*bufptr == '\n'))
      {
        /*
         * Use the last command entered - put 'last' back into
         * 'buffer' so it doesn't get destroyed by ProcessCommand()
         */
        strncpy(buffer, last, MAXLINE);
        cmdptr = buffer;
      }
      else
      {
        /*
         * Save this command for later use
         */
        strncpy(last, bufptr, MAXLINE);
        last[MAXLINE] = '\0';
        cmdptr = bufptr;
      }

      /*
       * Process given command
       */
      ProcessCommand(cmdptr);

      FreeLine(bufptr);
    } /* while (1) */
  }

#ifdef USE_CURSES

  if (ModeCurses)
  {
    assert(CurrentFrame != 0);

    while (1)
    {
      /*DisplayPrompt();*/

      RefreshFrames();

      wGets(CurrentFrame->window, buffer, MAXLINE);

      if (*buffer)
      {
        AddInput(buffer);
        ProcessCommand(buffer);
      }
    }
  } /* if (ModeCurses) */

#endif /* USE_CURSES */

} /* GetCommandInput() */

/*
FindCommand()
 Attempt to find the command "name" in the array "cmdlist"

Inputs: cmdlist - array to look in
        name    - command name
        flags   - bitmasks

Return: pointer to the index containing "name" if found,
        otherwise NULL.
        If the command is found, but there is more than 1 match
        (ambiguous), set the C_AMBIGUOUS bit in 'flags'
*/

struct Command *
FindCommand(struct Command *cmdlist, char *name, unsigned int *flags)

{
  struct Command *cmdptr, *tmp;
  int matches, /* number of matches we've had so far */
      clength;

  assert(name != 0);
  assert(cmdlist != 0);

  tmp = NULL;
  matches = 0;
  clength = strlen(name);
  for (cmdptr = cmdlist; cmdptr->cmd; ++cmdptr)
  {
    if (!Strncasecmp(name, cmdptr->cmd, clength))
    {
      if (clength == strlen(cmdptr->cmd))
      {
        /*
         * name and cmdptr->cmd are the same length, so it
         * must be an exact match, don't search any further
         */
        matches = 0;
        break;
      }
      tmp = cmdptr;
      ++matches;
    }
  }

  /*
   * If matches > 1, name is an ambiguous command, so the
   * user needs to be more specific
   */
  if ((matches == 1) && (tmp))
    cmdptr = tmp;

  if (cmdptr->cmd)
    return (cmdptr);

  if (matches != 0)
    *flags |= C_AMBIGUOUS;

  return (0);
} /* FindCommand() */
